from typing import Optional, List, Tuple, Any, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from deepchem.feat.molecule_featurizers.molgan_featurizer import GraphMatrix
from deepchem.models.torch_models import WGANModel
from deepchem.models.torch_models.layers import MolGANEncoderLayer


class BasicMolGANModel(WGANModel):
    """
    Model for de-novo generation of small molecules based on work of Nicola De Cao et al. [1]_.
    It uses a GAN directly on graph data and a reinforcement learning objective to induce the network to generate molecules with certain chemical properties.
    Utilizes WGAN infrastructure; uses adjacency matrix and node features as inputs.
    Inputs need to be one-hot representation.

    Examples
    --------

    Import necessary libraries and modules

    >>> import deepchem as dc
    >>> from deepchem.models.torch_models import BasicMolGANModel as MolGAN
    >>> from deepchem.models.optimizers import ExponentialDecay
    >>> import torch
    >>> import torch.nn.functional as F

    Load dataset and featurize molecules
    We will use a small dataset for this example.
    We will be using `MolGanFeaturizer` to featurize the molecules.

    >>> smiles = ['CCC', 'C1=CC=CC=C1', 'CNC' ]
    >>> # create featurizer
    >>> feat = dc.feat.MolGanFeaturizer()
    >>> # featurize molecules
    >>> features = feat.featurize(smiles)
    >>> # Remove empty objects
    >>> features = list(filter(lambda x: x is not None, features))

    Create and train the model

    >>> # create model
    >>> gan = MolGAN(learning_rate=ExponentialDecay(0.001, 0.9, 5000))
    >>> dataset = dc.data.NumpyDataset([x.adjacency_matrix for x in features],[x.node_features for x in features])
    >>> def iterbatches(epochs):
    ...     for i in range(epochs):
    ...         for batch in dataset.iterbatches(batch_size=gan.batch_size, pad_batches=True):
    ...             adjacency_tensor = F.one_hot(
    ...                     torch.Tensor(batch[0]).to(torch.int64),
    ...                     gan.edges).to(torch.float32)
    ...             node_tensor = F.one_hot(
    ...                     torch.Tensor(batch[1]).to(torch.int64),
    ...                     gan.nodes).to(torch.float32)
    ...             yield {gan.data_inputs[0]: adjacency_tensor, gan.data_inputs[1]:node_tensor}
    >>> # train model
    >>> gan.fit_gan(iterbatches(8), generator_steps=0.2, checkpoint_interval=0)

    You can change the above parameters to get better results. The above example is just a simple example to show how to use the model.
    You can try `iterbatches(1000)` for better results.

    Now, let's generate some molecules using the trained model
    We will generate 10 molecules and then convert them to RDKit molecules.

    >>> generated_data = gan.predict_gan_generator(10)
    Generating 10 samples
    >>> # convert graphs to RDKitmolecules
    >>> nmols = feat.defeaturize(generated_data)
    >>> print("{} molecules generated".format(len(nmols)))
    10 molecules generated

    You can increase the number of generated molecules by changing the parameter in `predict_gan_generator` function.
    Generated molecules are in the form of GraphMatrix. You can convert them to RDKit molecules using `defeaturize` function of MolGanFeaturizer.

    Now, let's remove invalid molecules from the generated molecules.

    >>> # remove invalid moles
    >>> nmols = list(filter(lambda x: x is not None, nmols))
    >>> print ("{} valid molecules".format(len(nmols)))
    0 valid molecules

    We can see that currently training is unstable and 0 is a common outcome. You can try training the model with different parameters to get better results.

    References
    ----------
    .. [1] Nicola De Cao et al. "MolGAN: An implicit generative model
        for small molecular graphs", https://arxiv.org/abs/1805.11973
    """

    def __init__(self,
                 edges: int = 5,
                 vertices: int = 9,
                 nodes: int = 5,
                 embedding_dim: int = 10,
                 dropout_rate: float = 0.0,
                 device: Optional[torch.device] = None,
                 **kwargs):
        """
        Initialize the model

        Parameters
        ----------
        edges: int, default 5
            Number of bond types includes BondType.Zero
        vertices: int, default 9
            Max number of atoms in adjacency and node features matrices
        nodes: int, default 5
            Number of atom types in node features matrix
        embedding_dim: int, default 10
            Size of noise input array
        dropout_rate: float, default = 0.
            Rate of dropout used across whole model
        name: str, default ''
            Name of the model
        """

        self.edges = edges
        self.vertices = vertices
        self.nodes = nodes
        self.embedding_dim = embedding_dim
        self.dropout_rate = dropout_rate
        if device is None:
            if torch.cuda.is_available():
                self.device = torch.device('cuda')
            elif torch.backends.mps.is_available():
                self.device = torch.device('mps')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        super(BasicMolGANModel, self).__init__(device=device, **kwargs)

    def get_noise_input_shape(self) -> Tuple[int, int]:
        """
        Return shape of the noise input used in generator

        Returns
        -------
        Tuple
            Shape of the noise input
        """

        return (
            1,
            self.embedding_dim,
        )

    def get_data_input_shapes(self) -> List:
        """
        Return input shape of the discriminator

        Returns
        -------
        List
            List of shapes used as an input for distriminator.
        """
        return [
            (1, self.vertices, self.vertices, self.edges),
            (1, self.vertices, self.nodes),
        ]

    def create_generator(self):
        """
        Create generator model.
        Take noise data as an input and processes it through number of
        dense and dropout layers. Then data is converted into two forms
        one used for training and other for generation of compounds.
        The model has two outputs:
            1. edges
            2. nodes

        The format differs depending on intended use (training or sample generation).
        For sample generation use flag, sample_generation=True while calling generator
        i.e. gan.generators[0](noise_input, training=False, sample_generation=True).
        For training the model, set `sample_generation=False`
        """
        return BasicMolGANGenerator(vertices=self.vertices,
                                    edges=self.edges,
                                    nodes=self.nodes,
                                    dropout_rate=self.dropout_rate,
                                    embedding_dim=self.embedding_dim)

    def create_discriminator(self,
                             units: List[Union[Tuple[int, int],
                                               int]] = [(128, 64), 64]):
        """
        Create discriminator model based on MolGAN layers.
        Takes two inputs:
            1. adjacency tensor, containing bond information
            2. nodes tensor, containing atom information

        The input vectors need to be in one-hot encoding format.
        Use MolGAN featurizer for that purpose. It will be simplified
        in the future release.
        """

        return Discriminator(dropout_rate=self.dropout_rate,
                             units=units,
                             edges=self.edges,
                             nodes=self.nodes,
                             device=self.device)

    def predict_gan_generator(self,
                              batch_size: int = 1,
                              noise_input: Optional[Union[List,
                                                          torch.Tensor]] = None,
                              conditional_inputs: List = [],
                              generator_index: int = 0) -> List[GraphMatrix]:
        """
        Use the GAN to generate a batch of samples.

        Parameters
        ----------
        batch_size: int
            the number of samples to generate.  If either noise_input or
            conditional_inputs is specified, this argument is ignored since the batch
            size is then determined by the size of that argument.
        noise_input: array
            the value to use for the generator's noise input.  If None (the default),
            get_noise_batch() is called to generate a random input, so each call will
            produce a new set of samples.
        conditional_inputs: list of arrays
            NOT USED.
            the values to use for all conditional inputs.  This must be specified if
            the GAN has any conditional inputs.
        generator_index: int
            NOT USED.
            the index of the generator (between 0 and n_generators-1) to use for
            generating the samples.

        Returns
        -------
        List[GraphMatrix]
            Returns a list of GraphMatrix object that can be converted into
            RDKit molecules using MolGANFeaturizer defeaturize function.
        """
        if noise_input is not None:
            batch_size = len(noise_input)
        if noise_input is None:
            noise_input = self.get_noise_batch(batch_size)  # type: ignore
        noise_input = torch.tensor(noise_input,
                                   dtype=torch.float32,
                                   device=self.device)
        print(f"Generating {batch_size} samples")
        adjacency_matrix, nodes_features = self.generators[0](
            noise_input, training=False, sample_generation=True)
        graphs = [
            GraphMatrix(i, j)
            for i, j in zip(adjacency_matrix.cpu().detach().numpy(),
                            nodes_features.cpu().detach().numpy())
        ]
        return graphs


class BasicMolGANGenerator(nn.Module):
    """
    Generator class for BasicMolGAN model.
    Using subclassing rather than functional API due to requirement
    to swap between two outputs depending on situation.
    In order to get output that used for sample generation
    (conversion to rdkit molecules) pass sample_generation=True argument while
    calling the model i.e. adjacency_matrix, nodes_features = self.generators[0](
    noise_input, training=False, sample_generation=True)
    This is automatically done in predict_gan_generator().
    """

    def __init__(self,
                 vertices: int = 9,
                 edges: int = 5,
                 nodes: int = 5,
                 dropout_rate: float = 0.0,
                 embedding_dim: int = 10,
                 name: str = "SimpleMolGANGenerator",
                 **kwargs):
        """
        Initialize model.

        Parameters
        ----------
        vertices : int, optional
            number of max atoms dataset molecules (incl. empty atom), by default 9
        edges : int, optional
            number of bond types in molecules, by default 5
        nodes : int, optional
            number of atom types in molecules, by default 5
        dropout_rate : float, optional
            rate of dropout, by default 0.
        embedding_dim : int, optional
            noise input dimensions, by default 10
        name : str, optional
            name of the model, by default "SimpleMolGANGenerator"
        """
        super(BasicMolGANGenerator, self).__init__(**kwargs)
        self.vertices = vertices
        self.edges = edges
        self.nodes = nodes
        self.dropout_rate = dropout_rate
        self.embedding_dim = embedding_dim

        self.dense1 = nn.Linear(self.embedding_dim, 128)  # tanh
        self.dropout1 = nn.Dropout(dropout_rate)
        self.dense2 = nn.Linear(128, 256)  # tanh
        self.dropout2 = nn.Dropout(dropout_rate)
        self.dense3 = nn.Linear(256, 512)  # tanh
        self.dropout3 = nn.Dropout(dropout_rate)

        # edges logits used during training
        self.edges_dense = nn.Linear(512,
                                     self.edges * self.vertices * self.vertices)
        self.edges_dropout = nn.Dropout(dropout_rate)

        # nodes logits used during training
        self.nodes_dense = nn.Linear(512, self.vertices * self.nodes)
        self.nodes_dropout = nn.Dropout(self.dropout_rate)

    def forward(self,
                inputs: Any,
                training: bool = False,
                sample_generation: bool = False) -> List[Any]:
        """
        Call generator model

        Parameters
        ----------
        inputs : Any
            List of inputs, typically noise_batch
        training : bool, optional
            used by dropout layers, by default False
        sample_generation : bool, optional
            decide which output to use, by default False

        Returns
        -------
        List[Any, Any]
            Tensors containing either softmax values for training
            or argmax for sample generation (used for creation of rdkit molecules).
        """

        if isinstance(inputs, list):
            inputs = inputs[0]
        x = F.tanh(self.dense1(inputs))
        x = self.dropout1(x)
        x = F.tanh(self.dense2(x))
        x = self.dropout2(x)
        x = F.tanh(self.dense3(x))
        x = self.dropout3(x)

        # edges logits
        edges_logits = self.edges_dense(x)
        edges_logits = edges_logits.view(-1, self.edges, self.vertices,
                                         self.vertices)
        matrix_transpose = edges_logits.permute(0, 1, 3, 2)
        edges_logits = (edges_logits + matrix_transpose) / 2
        edges_logits = edges_logits.permute(0, 2, 3, 1)
        edges_logits = self.edges_dropout(edges_logits)

        # nodes logits
        nodes_logits = self.nodes_dense(x)
        nodes_logits = nodes_logits.view(-1, self.vertices, self.nodes)
        nodes_logits = self.nodes_dropout(nodes_logits)

        if sample_generation is False:
            # For training
            edges = F.softmax(edges_logits, dim=-1)
            nodes = F.softmax(nodes_logits, dim=-1)
        else:
            # For sample generation
            e_gumbel_logits = edges_logits - torch.log(-torch.log(
                torch.rand_like(edges_logits, dtype=edges_logits.dtype)))
            e_gumbel_argmax = F.one_hot(torch.argmax(e_gumbel_logits, dim=-1),
                                        num_classes=e_gumbel_logits.shape[-1])
            edges = torch.argmax(e_gumbel_argmax, dim=-1)

            n_gumbel_logits = nodes_logits - torch.log(-torch.log(
                torch.rand_like(nodes_logits, dtype=nodes_logits.dtype)))
            n_gumbel_argmax = F.one_hot(torch.argmax(n_gumbel_logits, dim=-1),
                                        num_classes=n_gumbel_logits.shape[-1])
            nodes = torch.argmax(n_gumbel_argmax, dim=-1)
        return [edges, nodes]


class Discriminator(nn.Module):
    """A discriminator for the MolGAN model."""

    def __init__(
        self,
        dropout_rate: float,
        units: List = [(128, 64), 64],
        edges: int = 5,
        nodes: int = 5,
        device: Optional[torch.device] = torch.device('cpu')
    ) -> None:
        """Initialize the discriminator.

        Parameters
        ----------
        dropout_rate : float
            Rate of dropout used across whole model
        units : List, optional
            Units for MolGAN encoder layer, by default [(128, 64), 64]
        edges : int, optional
            Edge types, by default 5
        device : Optional[torch.device], optional
            Device to use, by default torch.device('cpu')
        """
        super(Discriminator, self).__init__()
        self.dropout_rate = dropout_rate
        self.edges = edges
        self.units = units
        self.nodes = nodes
        self.device: torch.device = device  # type: ignore
        self.graph = MolGANEncoderLayer(units=self.units,
                                        dropout_rate=self.dropout_rate,
                                        edges=self.edges,
                                        nodes=self.nodes,
                                        device=self.device)

        # Define the dense layers
        self.dense1 = nn.Linear(units[1], 128)
        self.dropout1 = nn.Dropout(dropout_rate)
        self.dense2 = nn.Linear(128, 64)
        self.dropout2 = nn.Dropout(dropout_rate)
        self.dense3 = nn.Linear(64, 1)

    def forward(
        self,
        inputs: List[torch.Tensor],
    ) -> torch.Tensor:
        """Forward pass for the discriminator.

        Parameters
        ----------
        inputs : List[torch.Tensor]
            List of inputs, typically adjacency_tensor and node_tensor

        Returns
        -------
        torch.Tensor
            Output tensor of the discriminator
        """
        adjacency_tensor, node_tensor = inputs

        if isinstance(adjacency_tensor, list):
            adjacency_tensor = adjacency_tensor[0]
        adjacency_tensor = adjacency_tensor.to(device=self.device,
                                               dtype=torch.float32)
        node_tensor = node_tensor.to(device=self.device, dtype=torch.float32)

        graph = self.graph([adjacency_tensor, node_tensor])

        graph = graph.to(device=self.device, dtype=torch.float32)
        output = self.dense1(graph)
        output = F.tanh(output)
        output = self.dropout1(output)
        output = self.dense2(output)
        output = F.tanh(output)
        output = self.dropout2(output)
        output = self.dense3(output)
        return output



from deepchem.molnet.load_function.bace_datasets import load_bace_classification, load_bace_regression
from deepchem.molnet.load_function.bbbc_datasets import load_bbbc001, load_bbbc002, load_bbbc003, load_bbbc004, load_bbbc005
from deepchem.molnet.load_function.bbbp_datasets import load_bbbp
from deepchem.molnet.load_function.cell_counting_datasets import load_cell_counting
from deepchem.molnet.load_function.chembl_datasets import load_chembl
from deepchem.molnet.load_function.clearance_datasets import load_clearance
from deepchem.molnet.load_function.clintox_datasets import load_clintox
from deepchem.molnet.load_function.delaney_datasets import load_delaney
from deepchem.molnet.load_function.hiv_datasets import load_hiv
from deepchem.molnet.load_function.hopv_datasets import load_hopv
from deepchem.molnet.load_function.kaggle_datasets import load_kaggle
from deepchem.molnet.load_function.lipo_datasets import load_lipo
from deepchem.molnet.load_function.muv_datasets import load_muv
from deepchem.molnet.load_function.nci_datasets import load_nci
from deepchem.molnet.load_function.pcba_datasets import load_pcba
from deepchem.molnet.load_function.pdbbind_datasets import load_pdbbind
from deepchem.molnet.load_function.ppb_datasets import load_ppb
from deepchem.molnet.load_function.qm7_datasets import load_qm7
from deepchem.molnet.load_function.qm8_datasets import load_qm8
from deepchem.molnet.load_function.qm9_datasets import load_qm9
from deepchem.molnet.load_function.sampl_datasets import load_sampl
from deepchem.molnet.load_function.sider_datasets import load_sider
from deepchem.molnet.load_function.sweetlead_datasets import load_sweet
from deepchem.molnet.load_function.tox21_datasets import load_tox21
from deepchem.molnet.load_function.toxcast_datasets import load_toxcast
from deepchem.molnet.load_function.uspto_datasets import load_uspto
from deepchem.molnet.load_function.uv_datasets import load_uv
from deepchem.molnet.load_function.factors_datasets import load_factors
from deepchem.molnet.load_function.kinase_datasets import load_kinase
from deepchem.molnet.load_function.thermosol_datasets import load_thermosol
from deepchem.molnet.load_function.hppb_datasets import load_hppb
from deepchem.molnet.load_function.chembl25_datasets import load_chembl25
from deepchem.molnet.load_function.zinc15_datasets import load_zinc15
from deepchem.molnet.load_function.freesolv_dataset import load_freesolv
from deepchem.molnet.load_function.material_datasets.load_bandgap import load_bandgap
from deepchem.molnet.load_function.material_datasets.load_perovskite import load_perovskite
from deepchem.molnet.load_function.material_datasets.load_Pt_NO_surface_adsorbate_energy import load_Platinum_Adsorption

from deepchem.molnet.load_function.material_datasets.load_mp_formation_energy import load_mp_formation_energy
from deepchem.molnet.load_function.material_datasets.load_mp_metallicity import load_mp_metallicity

from deepchem.molnet.load_function.molnet_loader import featurizers, splitters, transformers, TransformerGenerator, _MolnetLoader

from deepchem.molnet.dnasim import simulate_motif_density_localization
from deepchem.molnet.dnasim import simulate_motif_counting
from deepchem.molnet.dnasim import simple_motif_embedding
from deepchem.molnet.dnasim import motif_density
from deepchem.molnet.dnasim import simulate_single_motif_detection



from abc import abstractmethod
from typing import Union, List
import torch
from dqc.utils.datastruct import ValGrad
from deepchem.utils.dftutils import SpinParam
from dqc.api.getxc import get_xc
from dqc.xc.base_xc import BaseXC
from dqc.utils.safeops import safenorm, safepow


class BaseNNXC(BaseXC, torch.nn.Module):
    """
    Base class for the Neural Network XC (NNXC)  and HybridXC classes.

    Density-functional theory (DFT) is a theory used to calculate the
    electronic structure of atoms, molecules, and solids. Its objective is
    to use the fundamental laws of quantum mechanics to quantitatively
    comprehend the properties of materials.

    There are serious limitations to the tradional methods used to approximate
    solutions to the Schrödinger equation of N interacting electrons moving in
    an external potential. Whereas in DFT, instead of the many-body wave
    function, the density (n(r)) is a function of three spatial coordinates.

    The many-body electronic ground state can be described using single-particle    equations and an effective potential thanks to the Kohn-Sham theory. The
    exchange-correlation potential, which accounts for many-body effects, the
    Hartree potential, which describes the electrostatic electron-electron
    interaction, and the ionic potential resulting from the atomic cores make
    up the effective potential.

    The difference between the total exact energy and the total of the rest
    of the energy terms (such as kinetic energy), is known as the
    exchange-correlation energy. The exchange-correlation functional is obtained    by calculating the functional derivate of the XC energy w.r.t the
    electron density function. In this model, we are trying to build a neural
    network that can be trained to calculate an exchange-correlation functional
    based on a specific set of molecules/atoms/ions.

    This base class can be used to build layers such as the NNLDA layer, where
    the exchange correlation functional is trained based on the pre-defined LDA     class of functionals. The methods in this class take the electron density as
    the input and transform it accordingly. For example; The NNLDA layer
    requires only the density to build an NNXC whereas a GGA based model would
    require the density gradient as well. This method also takes polarization
    into account.

    References
    ----------
    Encyclopedia of Condensed Matter Physics, 2005.
    Mark R. Pederson, Tunna Baruah, in Advances In Atomic, Molecular, and
    Optical Physics, 2015
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation
    functional from nature with fully differentiable density functional
    theory." Physical Review Letters 127.12 (2021): 126403.
    """

    @abstractmethod
    def get_edensityxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]) -> torch.Tensor:
        """
        This method is used to transform the electron density. The output
        of this method varies depending on the layer.

        Parameters
        ----------
        densinfo: Union[ValGrad, SpinParam[ValGrad]]
            Density information calculated using DQC utilities.
        """
        pass

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        This method is implemented only to avoid errors while passing the
        get_edensityxc method values to DQC and Xitorch.
        """
        # torch.nn.module prefix has no ending dot, while xt prefix has
        nnprefix = prefix if prefix == "" else prefix[:-1]
        return [
            name for (name, param) in self.named_parameters(prefix=nnprefix)
        ]


class NNLDA(BaseNNXC):
    """
    Neural network xc functional of LDA

    Local-density approximations (LDA) are a class of approximations to the
    exchange–correlation (XC) energy. The LDA assumes variations of the
    density to be gradual, i.e, it is  based on the homogeneous electron
    gas model. Which is why it is regarded as the simplest approach to the
    exchange correlation functional. This class of functionals depend only
    upon the value of the electronic density at each point in space.

    Hence, in this model, we only input the density and not other components
    such as the gradients of the density (which is used in other functionals
    such as the GGA class).

    Examples
    --------
    >>> from deepchem.models.dft.nnxc import NNLDA
    >>> import torch
    >>> import torch.nn as nn
    >>> n_input, n_hidden = 2, 1
    >>> nnmodel = (nn.Linear(n_input, n_hidden))
    >>> output = NNLDA(nnmodel)

    References
    ----------
    Density-Functional Theory of the Electronic Structure of Molecules,Robert
    G. Parr and Weitao Yang, Annual Review of Physical Chemistry 1995 46:1,
    701-728
    R. O. Jones and O. Gunnarsson, Rev. Mod. Phys. 61, 689 (1989)
    """

    def __init__(self, nnmodel: torch.nn.Module):
        super().__init__()
        """
        Parameters
        ----------
        nnmodel: torch.nn.Module
            Neural network for xc functional
        """
        self.nnmodel = nnmodel

    def get_edensityxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]) -> torch.Tensor:
        """
        This method transform the local electron density (n) and the spin
        density (xi) for polarized and unpolarized cases, to be the input of
        the neural network.

        Parameters
        ----------
        densinfo: Union[ValGrad, SpinParam[ValGrad]]
            Density information calculated using DQC utilities.

        Returns
        -------
        res
            Neural network output by calculating total density (n) and the spin
            density (xi). The shape of res is (ninp , ) where ninp is the number            of layers in nnmodel ; which is user defined.
        """
        if isinstance(densinfo, ValGrad):  # unpolarized case
            n = densinfo.value.unsqueeze(-1)  # (*BD, nr, 1)
            xi = torch.zeros_like(n)
        else:  # polarized case
            nu = densinfo.u.value.unsqueeze(-1)
            nd = densinfo.d.value.unsqueeze(-1)
            n = nu + nd  # (*BD, nr, 1)
            xi = (nu - nd) / (n + 1e-18)  # avoiding nan

        ninp = n

        x = torch.cat((ninp, xi), dim=-1)  # (*BD, nr, 2)
        nnout = self.nnmodel(x)  # (*BD, nr, 1)
        res = nnout * n  # (*BD, nr, 1)
        res = res.squeeze(-1)
        return res


class NNPBE(BaseNNXC):
    # neural network xc functional of GGA (receives the density and grad as inputs)
    """
    Neural network xc functional of GGA class of functionals

    The Purdew-Burke-Ernzerhof (PBE) functional is a popular non-empirical
    functional which falls under the Generalized Gradient Approximation (GGA)
    class of functionals. GGA differs from LDA by incorporating nonlocal
    corrections involving gradients (or derivatives) of the electron density.
    In other words, the XC potential is a complicated function in 3D space
    since it depends on the electron density and its gradient.

    Hence, in this model we input the electron density as well as the
    calculated gradients. We calculate these values using the DQC and Xitorch
    libraries.

    Examples
    --------
    >>> from deepchem.models.dft.nnxc import NNPBE
    >>> import torch
    >>> import torch.nn as nn
    >>> n_input, n_hidden = 3, 3
    >>> nnmodel = (nn.Linear(n_input, n_hidden))
    >>> output = NNPBE(nnmodel)

    References
    ----------
    Perdew, John P., Kieron Burke, and Matthias Ernzerhof. "Generalized gradient approximation made simple." Physical review letters 77.18 (1996): 3865.
    A.W. Ghosh,
    5.09 - Electronics with Molecules,
    Editor(s): Pallab Bhattacharya, Roberto Fornari, Hiroshi Kamimura,
    Comprehensive Semiconductor Science and Technology,
    Elsevier,
    2011,
    Pages 383-479,
    ISBN 9780444531537,
    https://doi.org/10.1016/B978-0-44-453153-7.00033-X.
    """

    def __init__(self, nnmodel: torch.nn.Module):
        """
        Parameters
        ----------
        nnmodel: torch.nn.Module
            Neural network for xc functional. Shape; (3,...). This is because
            the nnmodel receives an input with the shape (....,3). This
            dimension is for;
            (0) total density (n): (n_up + n_dn), and
            (1) spin density (xi): (n_up - n_dn) / (n_up + n_dn)
            (2) normalized gradients (s): |del(n)| / [2(3*pi^2)^(1/3) * n^(4/3)]
        """
        super().__init__()
        self.nnmodel = nnmodel

    def get_edensityxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]) -> torch.Tensor:
        """
        This method transform the local electron density (n), its gradient
        and the spin density (xi) for polarized and unpolarized cases, to be
        the input of the neural network.

        Parameters
        ----------
        densinfo: Union[ValGrad, SpinParam[ValGrad]]
            Density information calculated using DQC utilities.

        Returns
        -------
        res
            Neural network output by calculating total density (n) and the spin
            density (xi). The shape of res is (ninp , ) where ninp is the number            of layers in nnmodel ; which is 3 for NNPBE. The shape of the output is (....,1) and it represents the energy density per density per unit volume.
        """
        # densinfo.value: (*BD, nr)
        # densinfo.grad : (*BD, nr, 3)

        # collect the total density (n), spin density (xi), and normalized gradients (s)
        a = 6.187335452560271  # 2 * (3 * np.pi ** 2) ** (1.0 / 3)
        if isinstance(densinfo, ValGrad):  # unpolarized case
            assert densinfo.grad is not None
            n = densinfo.value.unsqueeze(-1)  # (*BD, nr, 1)
            xi = torch.zeros_like(n)
            n_offset = n + 1e-18  # avoiding nan
            s = safenorm(densinfo.grad, dim=0).unsqueeze(-1)
        else:  # polarized case
            assert densinfo.u.grad is not None
            assert densinfo.d.grad is not None
            nu = densinfo.u.value.unsqueeze(-1)
            nd = densinfo.d.value.unsqueeze(-1)
            n = nu + nd  # (*BD, nr, 1)
            n_offset = n + 1e-18  # avoiding nan
            xi = (nu - nd) / n_offset
            s = safenorm(densinfo.u.grad + densinfo.d.grad, dim=0).unsqueeze(-1)

        # normalize the gradient
        s = s / a * safepow(n, -4.0 / 3)

        # decide how to transform the density to be the input of nn
        ninp = n
        sinp = s

        # get the neural network output
        x = torch.cat((ninp, xi, sinp), dim=-1)  # (*BD, nr, 3)
        nnout = self.nnmodel(x)  # (*BD, nr, 1)
        res = nnout * n
        res = res.squeeze(-1)
        return res


class HybridXC(BaseNNXC):
    """
    The HybridXC module computes XC energy by summing XC energy computed
    from libxc(any conventional DFT functional) and the trainable neural
    network with tunable weights.
    This layer constructs a hybrid functional based on the user's choice
    of what model is to be used to train the functional. (Currently, we only
    support an LDA based model). This hybrid functional is a combination of
    the xc that is trained by a neural network, and a conventional DFT
    functional.

    Examples
    --------
    >>> from deepchem.models.dft.nnxc import HybridXC
    >>> import torch
    >>> import torch.nn as nn
    >>> n_input, n_hidden = 2, 1
    >>> nnmodel = (nn.Linear(n_input, n_hidden))
    >>> output = HybridXC("lda_x", nnmodel, aweight0=0.0)
    """

    def __init__(self,
                 xcstr: str,
                 nnmodel: torch.nn.Module,
                 aweight0: float = 0.0,
                 bweight0: float = 1.0):

        super().__init__()
        """
        Parameters
        ----------
        xcstr: str
            The choice of xc to use. Some of the commonly used ones are:
            lda_x, lda_c_pw, lda_c_ow, lda_c_pz, lda_xc_lp_a, lda_xc_lp_b.
            The rest of the possible values can be found under the
            "LDA Functionals" section in the reference given below.
        nnmodel: nn.Module
            trainable neural network for prediction xc energy.
        aweight0: float
            weight of the neural network
        bweight0: float
            weight of the default xc

        References
        ----------
        https://tddft.org/programs/libxc/functionals/
        """
        self.xc = get_xc(xcstr)
        family = self.xc.family
        if family == 1:
            self.nnxc = NNLDA(nnmodel)
        else:
            self.nnxc = NNPBE(nnmodel)
        self.aweight = torch.nn.Parameter(
            torch.tensor(aweight0, requires_grad=True))
        self.bweight = torch.nn.Parameter(
            torch.tensor(bweight0, requires_grad=True))
        self.weight_activation = torch.nn.Identity()

    @property
    def family(self) -> int:
        """
        This method determines the type of model to be used, to train the
        neural network. Currently we only support an LDA based model and will
        implement more in subsequent iterations.
        Returns
        -------
        xc.family
        """
        return self.xc.family

    def get_edensityxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]) -> torch.Tensor:
        """Get electron density from xc
        This function reflects eqn. 4 in the `paper <https://arxiv.org/abs/2102.04229>_`.
        Parameters
        ----------
        densinfo: Union[ValGrad, SpinParam[ValGrad]]
            Density information calculated using DQC utilities.

        Returns
        -------
        Total calculated electron density with tunable weights.
        """
        nnxc_ene = self.nnxc.get_edensityxc(densinfo)
        xc_ene = self.xc.get_edensityxc(densinfo)
        aweight = self.weight_activation(self.aweight)
        bweight = self.weight_activation(self.bweight)
        return nnxc_ene * aweight + xc_ene * bweight



from __future__ import annotations
from abc import abstractmethod
from typing import Union
import torch
import numpy as np
from dqc.qccalc.ks import KS
from deepchem.feat.dft_data import DFTEntry, DFTSystem
from deepchem.utils.dftutils import KSCalc, hashstr, SpinParam
from deepchem.models.dft.nnxc import BaseNNXC, HybridXC


class XCNNSCF(torch.nn.Module):
    """
    Exchange Correlation Neural Network - Self Consistent Iterations

    In the Kohn-Sham theory, the inclusion of the noninteracting kinetic energy     functional results in a set of one-particle equations with Kohn-Sham
    orbitals as their solutions after functional differentiation. It is a
    variational approach that determines the lowest energy and the related
    molecular orbitals and orbital energies by using the electron-electron
    interaction potential. To learn more about Density Functional Theory
    and the Kohn-Sham approach please use the references below.

    The XCNNSCF is used for performing self-consistent iterations. The
    XC functional in the Kohn-Sham model implementation is replaced by a
    neural network.

    Examples
    --------
    >>> from deepchem.models.dft.scf import XCNNSCF
    >>> import torch
    >>> from deepchem.feat.dft_data import DFTEntry, DFTSystem
    >>> from deepchem.models.dft.nnxc import HybridXC
    >>> nnmodel = (torch.nn.Sequential(
    ...         torch.nn.Linear(2, 10),
    ...         torch.nn.Tanh(),
    ...         torch.nn.Linear(10, 1))).to(torch.double)
    >>> hybridxc = HybridXC("lda_x", nnmodel, aweight0=0.0)
    >>> e_type = 'dm'
    >>> true_val = 'deepchem/feat/tests/data/dftHF_output.npy'
    >>> systems = [{
    ...     'moldesc': 'H 0.86625 0 0; F -0.86625 0 0',
    ...     'basis': '6-311++G(3df,3pd)'
    ... }]
    >>> entry = DFTEntry.create(e_type, true_val, systems)
    >>> evl = XCNNSCF(hybridxc, entry)
    >>> system = DFTSystem(systems[0])
    >>> run = evl.run(system)
    The 6-311++G(3df,3pd) basis for atomz 9 does not exist, but we will download it
    Downloaded to /home/runner/miniconda3/envs/deepchem/lib/python3.8/site-packages/dqc/api/.database/6-311ppg_3df_3pd_/09.gaussian94
    >>> output = run.energy()

    Notes
    -----
    This code is derived from https://github.com/mfkasim1/xcnn/blob/f2cb9777da2961ac553f256ecdcca3e314a538ca/xcdnn2/evaluator.py

    References
    ----------
    deepchem.models.dft.nnxc
    Kohn, W. and Sham, L.J., 1965. Self-consistent equations including
    exchange and correlation effects. Physical review, 140(4A), p.A1133.
    """

    def __init__(self, xc: Union[BaseNNXC, HybridXC], entry: DFTEntry):
        super().__init__()
        """
        Parameters
        ----------
        xc: Union[BaseNNXC, HybridXC]
            exchange correlation functional that has been replaced by a
            neural network.
        entry: DFTEntry
        """
        self.xc = xc

    @abstractmethod
    def get_xc(self) -> HybridXC:
        """
        Returns
        -------
        Exchange correlation functional that has been replaced by a
        neural network, based on a BaseNNXC model.
        """
        return self.xc

    @abstractmethod
    def run(self, system: DFTSystem) -> KSCalc:
        """
        Kohn Sham Model
        This method runs the Quantum Chemistry calculation (Differentiable
        DFT) of the given system and returns the post-run object. This method
        starts with an intial density matrix, the new density matrix can be
        obtained from the post-run object.

        Parameters
        ----------
        system: DFTSystem

        Returns
        -------
        KSCalc object

        """
        dm0, dmname = self._get_dm0(system)
        mol = system.get_dqc_mol()
        qc = KS(mol, xc=self.xc).run(dm0=dm0, bck_options={"max_niter": 50})
        return KSCalc(qc)

    def _dm0_name(self, obj) -> str:
        """
        Returns
        -------
        dm0 followed by the name of the system
        """
        return "dm0_" + hashstr(str(obj))

    def _get_dm0(self, system: DFTSystem):
        """
        This method calculates and retuns the density matrix of a system.
        The matrix will vary depending on the atomic numbers, positions, and
        spins.

        Parameters
        ----------
        system: DFTSystem
        """
        dm_name = self._dm0_name(system)
        dm0: torch.Tensor
        get_dm = np.array(getattr(self, dm_name, None), dtype=bool)
        dm0 = torch.Tensor(get_dm)
        dm_exists = dm0 is not None
        dm_written = dm_exists and torch.any(dm0 != 0.0)
        if not dm_written:
            dm0_res: Union[None, torch.Tensor, SpinParam[torch.Tensor]] = None
        elif system.get_dqc_mol().spin != 0:
            dm0_res = SpinParam(u=dm0[0].detach(), d=dm0[1].detach())
        else:
            dm0_res = dm0

        return dm0_res, (dm_name if dm_exists else None)



import logging
from typing import List
import numpy as np
import scipy
from deepchem.feat.base_classes import Featurizer
from typing import Any, Iterable

logger = logging.getLogger(__name__)
CHARSET = [
    'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R',
    'S', 'T', 'V', 'W', 'Y', 'X', 'Z', 'B', 'U', 'O'
]


class SparseMatrixOneHotFeaturizer(Featurizer):
    """Encodes any arbitrary string as a one-hot array.

    This featurizer uses the sklearn OneHotEncoder to create
    sparse matrix representation of a one-hot array of any string.
    It is expected to be used in large datasets that produces memory overload
    using standard featurizer such as OneHotFeaturizer. For example: SwissprotDataset


    Examples
    --------
    >>> import deepchem as dc
    >>> featurizer = dc.feat.SparseMatrixOneHotFeaturizer()
    >>> sequence = "MMMQLA"
    >>> encodings = featurizer.featurize([sequence])
    >>> encodings[0].shape
    (6, 25)

    """

    def __init__(self, charset: List[str] = CHARSET):
        """Initialize featurizer.

        Parameters
        ----------
        charset: List[str] (default code)
            A list of strings, where each string is length 1 and unique.

        """
        if len(charset) != len(set(charset)):
            raise ValueError("All values in charset must be unique.")
        self.charset = charset
        from sklearn.preprocessing import OneHotEncoder
        cat = np.array(self.charset).reshape(1, len(self.charset))
        self.ohe = OneHotEncoder(categories=list(cat), handle_unknown='ignore')

    def featurize(self,
                  datapoints: Iterable[Any],
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """Featurize strings.

        Parameters
        ----------
        datapoints: list
            A list of either strings (str or numpy.str_)
        log_every_n: int, optional (default 1000)
            How many elements are featurized every time a featurization is logged.

        """
        # Featurize data using featurize() in parent class
        return Featurizer.featurize(self, datapoints, log_every_n)

    def _featurize(self, datapoint: Any, **kwargs):
        """ Use parent method of base clase Featurizer.

        Parameters
        ----------
        datapoint : list of string
            string to be converted to a sparse one hot matrix.

        Returns
        -------
        scipy sparse matrix
            A scipy sparse matrix of the one hot representation of the given string.

        """
        # Featurize str data
        if isinstance(datapoint, (str, np.str_)):
            sequence = np.array(list(datapoint)).reshape(-1, 1)
            sparse_mat = self.ohe.fit_transform(sequence)
            return sparse_mat
        else:
            raise ValueError("Datapoint is not a string")

    def untransform(self, one_hot_vectors: scipy.sparse.spmatrix) -> str:
        """Convert from one hot representation back to original string

        Parameters
        ----------
        one_hot_vectors: np.ndarray
            An array of one hot encoded features.

        Returns
        -------
        str
            Original string for an one hot encoded array.

        """
        string = ""
        invers_trans = self.ohe.inverse_transform(one_hot_vectors)
        for one_hot in invers_trans:
            string += one_hot[0]
        return string



"""PyTorch implementation of fully connected networks.
"""
import logging
import numpy as np
import torch
import torch.nn.functional as F
from collections.abc import Sequence as SequenceCollection

import deepchem as dc
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import _make_pytorch_shapes_consistent
from deepchem.metrics import to_one_hot

from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Union
from deepchem.utils.typing import ActivationFn, LossFn, OneOrMany
from deepchem.utils.pytorch_utils import get_activation

logger = logging.getLogger(__name__)


class MultitaskClassifier(TorchModel):
    """A fully connected network for multitask classification.

    This class provides lots of options for customizing aspects of the model: the
    number and widths of layers, the activation functions, regularization methods,
    etc.

    It optionally can compose the model from pre-activation residual blocks, as
    described in https://arxiv.org/abs/1603.05027, rather than a simple stack of
    dense layers.  This often leads to easier training, especially when using a
    large number of layers.  Note that residual blocks can only be used when
    successive layers have the same width.  Wherever the layer width changes, a
    simple dense layer will be used even if residual=True.
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: Sequence[int] = [1000],
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = 'l2',
                 dropouts: OneOrMany[float] = 0.5,
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 n_classes: int = 2,
                 residual: bool = False,
                 **kwargs) -> None:
        """Create a MultitaskClassifier.

        In addition to the following arguments, this class also accepts
        all the keyword arguments from TensorGraph.

        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: int
            number of features
        layer_sizes: list
            the size of each dense layer in the network.  The length of
            this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight
            initialization of each layer.  The length of this list should
            equal len(layer_sizes).  Alternatively this may be a single
            value instead of a list, in which case the same value is used
            for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer to.  The
            length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in
            which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the PyTorch activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.  Standard activation functions from torch.nn.functional can be specified by name.
        n_classes: int
            the number of classes
        residual: bool
            if True, the model will be composed of pre-activation residual blocks instead
            of a simple stack of dense layers.
        """
        self.n_tasks = n_tasks
        self.n_features = n_features
        self.n_classes = n_classes
        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if isinstance(
                activation_fns,
                str) or not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers
        activation_fns = [get_activation(f) for f in activation_fns]

        # Define the PyTorch Module that implements the model.

        class PytorchImpl(torch.nn.Module):

            def __init__(self):
                super(PytorchImpl, self).__init__()
                self.layers = torch.nn.ModuleList()
                prev_size = n_features
                for size, weight_stddev, bias_const in zip(
                        layer_sizes, weight_init_stddevs, bias_init_consts):
                    layer = torch.nn.Linear(prev_size, size)
                    torch.nn.init.normal_(layer.weight, 0, weight_stddev)
                    torch.nn.init.constant_(layer.bias, bias_const)
                    self.layers.append(layer)
                    prev_size = size
                self.output_layer = torch.nn.Linear(prev_size,
                                                    n_tasks * n_classes)
                torch.nn.init.xavier_uniform_(self.output_layer.weight)
                torch.nn.init.constant_(self.output_layer.bias, 0)

            def forward(self, x):
                prev_size = n_features
                next_activation = None
                for size, layer, dropout, activation_fn, in zip(
                        layer_sizes, self.layers, dropouts, activation_fns):
                    y = x
                    if next_activation is not None:
                        y = next_activation(x)
                    y = layer(y)
                    if dropout > 0.0 and self.training:
                        y = F.dropout(y, dropout)
                    if residual and prev_size == size:
                        y = x + y
                    x = y
                    prev_size = size
                    next_activation = activation_fn
                if next_activation is not None:
                    y = next_activation(y)
                neural_fingerprint = y
                y = self.output_layer(y)
                logits = torch.reshape(y, (-1, n_tasks, n_classes))
                output = F.softmax(logits, dim=2)
                return (output, logits, neural_fingerprint)

        model = PytorchImpl()
        regularization_loss: Optional[Callable]
        if weight_decay_penalty != 0:
            weights = [layer.weight for layer in model.layers]
            if weight_decay_penalty_type == 'l1':
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.abs(w).sum() for w in weights]))
            else:
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.square(w).sum() for w in weights]))
        else:
            regularization_loss = None
        super(MultitaskClassifier,
              self).__init__(model,
                             dc.models.losses.SoftmaxCrossEntropy(),
                             output_types=['prediction', 'loss', 'embedding'],
                             regularization_loss=regularization_loss,
                             **kwargs)

    def default_generator(
            self,
            dataset: dc.data.Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                        -1, self.n_tasks, self.n_classes)
                yield ([X_b], [y_b], [w_b])


class MultitaskRegressor(TorchModel):
    """A fully connected network for multitask regression.

    This class provides lots of options for customizing aspects of the model: the
    number and widths of layers, the activation functions, regularization methods,
    etc.

    It optionally can compose the model from pre-activation residual blocks, as
    described in https://arxiv.org/abs/1603.05027, rather than a simple stack of
    dense layers.  This often leads to easier training, especially when using a
    large number of layers.  Note that residual blocks can only be used when
    successive layers have the same width.  Wherever the layer width changes, a
    simple dense layer will be used even if residual=True.
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: Sequence[int] = [1000],
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = 'l2',
                 dropouts: OneOrMany[float] = 0.5,
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 uncertainty: bool = False,
                 residual: bool = False,
                 **kwargs) -> None:
        """Create a MultitaskRegressor.

        In addition to the following arguments, this class also accepts all the keywork arguments
        from TensorGraph.

        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: int
            number of features
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes)+1.  The final element corresponds to the output layer.
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes)+1.
            The final element corresponds to the output layer.  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the PyTorch activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.  Standard activation functions from torch.nn.functional can be specified by name.
        uncertainty: bool
            if True, include extra outputs and loss terms to enable the uncertainty
            in outputs to be predicted
        residual: bool
            if True, the model will be composed of pre-activation residual blocks instead
            of a simple stack of dense layers.
        """
        self.n_tasks = n_tasks
        self.n_features = n_features
        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * (n_layers + 1)
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * (n_layers + 1)
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if isinstance(
                activation_fns,
                str) or not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers
        activation_fns = [get_activation(f) for f in activation_fns]
        if uncertainty:
            if any(d == 0.0 for d in dropouts):
                raise ValueError(
                    'Dropout must be included in every layer to predict uncertainty'
                )

        # Define the PyTorch Module that implements the model.

        class PytorchImpl(torch.nn.Module):

            def __init__(self):
                super(PytorchImpl, self).__init__()
                self.layers = torch.nn.ModuleList()
                prev_size = n_features
                for size, weight_stddev, bias_const in zip(
                        layer_sizes, weight_init_stddevs, bias_init_consts):
                    layer = torch.nn.Linear(prev_size, size)
                    torch.nn.init.normal_(layer.weight, 0, weight_stddev)
                    torch.nn.init.constant_(layer.bias, bias_const)
                    self.layers.append(layer)
                    prev_size = size
                self.output_layer = torch.nn.Linear(prev_size, n_tasks)
                torch.nn.init.normal_(self.output_layer.weight, 0,
                                      weight_init_stddevs[-1])
                torch.nn.init.constant_(self.output_layer.bias,
                                        bias_init_consts[-1])
                self.uncertainty_layer = torch.nn.Linear(prev_size, n_tasks)
                torch.nn.init.normal_(self.output_layer.weight, 0,
                                      weight_init_stddevs[-1])
                torch.nn.init.constant_(self.output_layer.bias, 0)

            def forward(self, inputs):
                x, dropout_switch = inputs
                prev_size = n_features
                next_activation = None
                for size, layer, dropout, activation_fn, in zip(
                        layer_sizes, self.layers, dropouts, activation_fns):
                    y = x
                    if next_activation is not None:
                        y = next_activation(x)
                    y = layer(y)
                    if dropout > 0.0 and dropout_switch:
                        y = F.dropout(y, dropout)
                    if residual and prev_size == size:
                        y = x + y
                    x = y
                    prev_size = size
                    next_activation = activation_fn
                if next_activation is not None:
                    y = next_activation(y)
                neural_fingerprint = y
                output = torch.reshape(self.output_layer(y), (-1, n_tasks, 1))
                if uncertainty:
                    log_var = torch.reshape(self.uncertainty_layer(y),
                                            (-1, n_tasks, 1))
                    var = torch.exp(log_var)
                    return (output, var, output, log_var, neural_fingerprint)
                else:
                    return (output, neural_fingerprint)

        model = PytorchImpl()
        regularization_loss: Optional[Callable]
        if weight_decay_penalty != 0:
            weights = [layer.weight for layer in model.layers]
            if weight_decay_penalty_type == 'l1':
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.abs(w).sum() for w in weights]))
            else:
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.square(w).sum() for w in weights]))
        else:
            regularization_loss = None
        loss: Union[dc.models.losses.Loss, LossFn]
        if uncertainty:
            output_types = [
                'prediction', 'variance', 'loss', 'loss', 'embedding'
            ]

            def loss(outputs, labels, weights):
                output, labels = _make_pytorch_shapes_consistent(
                    outputs[0], labels[0])
                diff = labels - output
                losses = diff * diff / torch.exp(outputs[1]) + outputs[1]
                w = weights[0]
                if len(w.shape) < len(losses.shape):
                    if isinstance(w, torch.Tensor):
                        shape = tuple(w.shape)
                    else:
                        shape = w.shape
                    shape = tuple(-1 if x is None else x for x in shape)
                    w = w.reshape(shape + (1,) *
                                  (len(losses.shape) - len(w.shape)))

                loss = losses * w
                loss = loss.mean()
                if regularization_loss is not None:
                    loss += regularization_loss()
                return loss
        else:
            output_types = ['prediction', 'embedding']
            loss = dc.models.losses.L2Loss()
        super(MultitaskRegressor,
              self).__init__(model,
                             loss,
                             output_types=output_types,
                             regularization_loss=regularization_loss,
                             **kwargs)

    def default_generator(
            self,
            dataset: dc.data.Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if mode == 'predict':
                    dropout = np.array(0.0)
                else:
                    dropout = np.array(1.0)
                yield ([X_b, dropout], [y_b], [w_b])


class MultitaskFitTransformRegressor(MultitaskRegressor):
    """Implements a MultitaskRegressor that performs on-the-fly transformation during fit/predict.

    Examples
    --------
    >>> n_samples = 10
    >>> n_features = 3
    >>> n_tasks = 1
    >>> ids = np.arange(n_samples)
    >>> X = np.random.rand(n_samples, n_features, n_features)
    >>> y = np.zeros((n_samples, n_tasks))
    >>> w = np.ones((n_samples, n_tasks))
    >>> dataset = dc.data.NumpyDataset(X, y, w, ids)
    >>> fit_transformers = [dc.trans.CoulombFitTransformer(dataset)]
    >>> model = dc.models.MultitaskFitTransformRegressor(n_tasks, [n_features, n_features],
    ...     dropouts=[0.], learning_rate=0.003, weight_init_stddevs=[np.sqrt(6)/np.sqrt(1000)],
    ...     batch_size=n_samples, fit_transformers=fit_transformers)
    >>> model.n_features
    12
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 fit_transformers: Sequence[dc.trans.Transformer] = [],
                 batch_size: int = 50,
                 **kwargs):
        """Create a MultitaskFitTransformRegressor.

        In addition to the following arguments, this class also accepts all the keywork arguments
        from MultitaskRegressor.

        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: list or int
            number of features
        fit_transformers: list
            List of dc.trans.FitTransformer objects
        """
        self.fit_transformers = fit_transformers

        # Run fit transformers on dummy dataset to determine n_features after transformation

        if isinstance(n_features, list):
            X_b = np.ones([batch_size] + n_features)
        elif isinstance(n_features, int):
            X_b = np.ones([batch_size, n_features])
        else:
            raise ValueError("n_features should be list or int")
        empty: np.ndarray = np.array([])
        for transformer in fit_transformers:
            assert transformer.transform_X and not (transformer.transform_y or
                                                    transformer.transform_w)
            X_b, _, _, _ = transformer.transform_array(X_b, empty, empty, empty)
        n_features = X_b.shape[1]
        logger.info("n_features after fit_transform: %d", int(n_features))
        super(MultitaskFitTransformRegressor,
              self).__init__(n_tasks,
                             n_features,
                             batch_size=batch_size,
                             **kwargs)

    def default_generator(
            self,
            dataset: dc.data.Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        empty: np.ndarray = np.array([])
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    y_b = y_b.reshape(-1, self.n_tasks, 1)
                if X_b is not None:
                    if mode == 'fit':
                        for transformer in self.fit_transformers:
                            X_b, _, _, _ = transformer.transform_array(
                                X_b, empty, empty, empty)
                if mode == 'predict':
                    dropout = np.array(0.0)
                else:
                    dropout = np.array(1.0)
                yield ([X_b, dropout], [y_b], [w_b])

    def predict_on_generator(
            self,
            generator: Iterable[Tuple[Any, Any, Any]],
            transformers: List[dc.trans.Transformer] = [],
            output_types: Optional[OneOrMany[str]] = None
    ) -> OneOrMany[np.ndarray]:

        def transform_generator():
            for inputs, labels, weights in generator:
                X_t = inputs[0]
                for transformer in self.fit_transformers:
                    X_t = transformer.X_transform(X_t)
                yield ([X_t] + inputs[1:], labels, weights)

        return super(MultitaskFitTransformRegressor,
                     self).predict_on_generator(transform_generator(),
                                                transformers, output_types)



# flake8: noqa
try:
    from deepchem.utils.differentiation_utils.editable_module import EditableModule

    from deepchem.utils.differentiation_utils.bcast import normalize_bcast_dims
    from deepchem.utils.differentiation_utils.bcast import get_bcasted_dims
    from deepchem.utils.differentiation_utils.bcast import match_dim

    from deepchem.utils.differentiation_utils.misc import set_default_option
    from deepchem.utils.differentiation_utils.misc import get_and_pop_keys
    from deepchem.utils.differentiation_utils.misc import get_method
    from deepchem.utils.differentiation_utils.misc import dummy_context_manager
    from deepchem.utils.differentiation_utils.misc import assert_runtime

    from deepchem.utils.differentiation_utils.linop import LinearOperator
    from deepchem.utils.differentiation_utils.linop import AddLinearOperator
    from deepchem.utils.differentiation_utils.linop import MulLinearOperator
    from deepchem.utils.differentiation_utils.linop import AdjointLinearOperator
    from deepchem.utils.differentiation_utils.linop import MatmulLinearOperator
    from deepchem.utils.differentiation_utils.linop import MatrixLinearOperator

    from deepchem.utils.differentiation_utils.pure_function import PureFunction
    from deepchem.utils.differentiation_utils.pure_function import get_pure_function
    from deepchem.utils.differentiation_utils.pure_function import make_sibling

    from deepchem.utils.differentiation_utils.grad import jac

    from deepchem.utils.differentiation_utils.solve import wrap_gmres
    from deepchem.utils.differentiation_utils.solve import exactsolve
    from deepchem.utils.differentiation_utils.solve import solve_ABE
    from deepchem.utils.differentiation_utils.solve import broyden1_solve
    from deepchem.utils.differentiation_utils.solve import get_batchdims
    from deepchem.utils.differentiation_utils.solve import setup_precond
    from deepchem.utils.differentiation_utils.solve import dot
    from deepchem.utils.differentiation_utils.solve import get_largest_eival
    from deepchem.utils.differentiation_utils.solve import safedenom
    from deepchem.utils.differentiation_utils.solve import setup_linear_problem
    from deepchem.utils.differentiation_utils.solve import gmres
    from deepchem.utils.differentiation_utils.solve import solve
    from deepchem.utils.differentiation_utils.solve import cg
    from deepchem.utils.differentiation_utils.solve import bicgstab

    from deepchem.utils.differentiation_utils.symeig import lsymeig
    from deepchem.utils.differentiation_utils.symeig import usymeig
    from deepchem.utils.differentiation_utils.symeig import symeig
    from deepchem.utils.differentiation_utils.symeig import ortho
    from deepchem.utils.differentiation_utils.symeig import exacteig
    from deepchem.utils.differentiation_utils.symeig import svd

    from deepchem.utils.differentiation_utils.optimize.rootsolver import broyden1
    from deepchem.utils.differentiation_utils.optimize.rootsolver import broyden2
    from deepchem.utils.differentiation_utils.optimize.rootsolver import linearmixing

    from deepchem.utils.differentiation_utils.optimize.equilibrium import anderson_acc

    from deepchem.utils.differentiation_utils.optimize.minimizer import gd
    from deepchem.utils.differentiation_utils.optimize.minimizer import adam

    from deepchem.utils.differentiation_utils.optimize.rootfinder import rootfinder
    from deepchem.utils.differentiation_utils.optimize.rootfinder import equilibrium
    from deepchem.utils.differentiation_utils.optimize.rootfinder import minimize
except:
    pass



def test_fourier_encode_dist():
    import numpy as np
    import torch

    from deepchem.utils.graph_utils import fourier_encode_dist
    x = torch.tensor([1.0, 2.0, 3.0])
    num_encodings = 4
    include_self = True

    encoded_x = fourier_encode_dist(x,
                                    num_encodings=num_encodings,
                                    include_self=include_self)
    assert encoded_x.shape == (x.shape[0],
                               num_encodings * 2 + int(include_self))

    scales = 2**np.arange(num_encodings)
    x_scaled = x.unsqueeze(-1) / scales
    x_sin = torch.sin(x_scaled)
    x_cos = torch.cos(x_scaled)
    x_expected = torch.cat([x_sin, x_cos], dim=-1)
    if include_self:
        x_expected = torch.cat((x_expected, x.unsqueeze(-1)), dim=-1)

    assert torch.allclose(encoded_x.float(), x_expected.float(), atol=1e-5)



"""
Atomic coordinate featurizer.
"""
import numpy as np

from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.utils.typing import RDKitMol


class AtomicCoordinates(MolecularFeaturizer):
    """Calculate atomic coordinates.

    Examples
    --------
    >>> import deepchem as dc
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('C1C=CC=CC=1')
    >>> n_atoms = len(mol.GetAtoms())
    >>> n_atoms
    6
    >>> featurizer = dc.feat.AtomicCoordinates(use_bohr=False)
    >>> features = featurizer.featurize([mol])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape # (n_atoms, 3)
    (6, 3)


    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self, use_bohr: bool = False):
        """
        Parameters
        ----------
        use_bohr: bool, optional (default False)
            Whether to use bohr or angstrom as a coordinate unit.

        """
        self.use_bohr = use_bohr

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Calculate atomic coordinates.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            A numpy array of atomic coordinates. The shape is `(n_atoms, 3)`.

        """
        try:
            from rdkit import Chem
            from rdkit.Chem import AllChem
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        # Check whether num_confs >=1 or not
        num_confs = len(datapoint.GetConformers())
        if num_confs == 0:
            datapoint = Chem.AddHs(datapoint)
            AllChem.EmbedMolecule(datapoint, AllChem.ETKDG())
            datapoint = Chem.RemoveHs(datapoint)

        N = datapoint.GetNumAtoms()
        coords = np.zeros((N, 3))

        # RDKit stores atomic coordinates in Angstrom. Atomic unit of length is the
        # bohr (1 bohr = 0.529177 Angstrom). Converting units makes gradient calculation
        # consistent with most QM software packages.
        if self.use_bohr:
            coords_list = [
                datapoint.GetConformer(0).GetAtomPosition(i).__idiv__(
                    0.52917721092) for i in range(N)
            ]
        else:
            coords_list = [
                datapoint.GetConformer(0).GetAtomPosition(i) for i in range(N)
            ]

        for atom in range(N):
            coords[atom, 0] = coords_list[atom].x
            coords[atom, 1] = coords_list[atom].y
            coords[atom, 2] = coords_list[atom].z

        return coords
